# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz mkdir.
"""

import os

from breezy import tests


class TestMkdir(tests.TestCaseWithTransport):

    def test_mkdir(self):
        tree = self.make_branch_and_tree('.')
        self.run_bzr(['mkdir', 'somedir'])
        self.assertEqual(tree.kind('somedir'), "directory")

    def test_mkdir_parents(self):
        tree = self.make_branch_and_tree('.')
        self.run_bzr(['mkdir', '-p', 'somedir/foo'])
        self.assertEqual(tree.kind('somedir/foo'), "directory")

    def test_mkdir_parents_existing_versioned_dir(self):
        tree = self.make_branch_and_tree('.')
        tree.mkdir('somedir')
        self.assertEqual(tree.kind('somedir'), "directory")
        self.run_bzr(['mkdir', '-p', 'somedir'])

    def test_mkdir_parents_existing_unversioned_dir(self):
        tree = self.make_branch_and_tree('.')
        os.mkdir('somedir')
        self.run_bzr(['mkdir', '-p', 'somedir'])
        self.assertEqual(tree.kind('somedir'), "directory")

    def test_mkdir_parents_with_unversioned_parent(self):
        tree = self.make_branch_and_tree('.')
        os.mkdir('somedir')
        self.run_bzr(['mkdir', '-p', 'somedir/foo'])
        self.assertEqual(tree.kind('somedir/foo'), "directory")



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for locking/unlocking a repository with external references."""

from breezy import branch, errors
from breezy.tests.per_repository_reference import \
    TestCaseWithExternalReferenceRepository


class TestUnlock(TestCaseWithExternalReferenceRepository):

    def create_stacked_branch(self):
        builder = self.make_branch_builder('source',
                                           format=self.bzrdir_format)
        builder.start_series()
        repo = builder.get_branch().repository
        if not repo._format.supports_external_lookups:
            raise tests.TestNotApplicable('format does not support stacking')
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'contents\n'))],
            revision_id=b'A-id')
        builder.build_snapshot([b'A-id'], [
            ('modify', ('file', b'new-content\n'))],
            revision_id=b'B-id')
        builder.build_snapshot([b'B-id'], [
            ('modify', ('file', b'yet more content\n'))],
            revision_id=b'C-id')
        builder.finish_series()
        source_b = builder.get_branch()
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        base = self.make_branch('base')
        base.pull(source_b, stop_revision=b'B-id')
        stacked = self.make_branch('stacked')
        stacked.set_stacked_on_url('../base')
        stacked.pull(source_b, stop_revision=b'C-id')

        return base, stacked

    def test_unlock_unlocks_fallback(self):
        base = self.make_branch('base')
        stacked = self.make_branch('stacked')
        repo = stacked.repository
        stacked.set_stacked_on_url('../base')
        self.assertEqual(1, len(repo._fallback_repositories))
        fallback_repo = repo._fallback_repositories[0]
        self.assertFalse(repo.is_locked())
        self.assertFalse(fallback_repo.is_locked())
        repo.lock_read()
        self.assertTrue(repo.is_locked())
        self.assertTrue(fallback_repo.is_locked())
        repo.unlock()
        self.assertFalse(repo.is_locked())
        self.assertFalse(fallback_repo.is_locked())
        repo.lock_write()
        self.assertTrue(repo.is_locked())
        self.assertTrue(fallback_repo.is_locked())
        repo.unlock()
        self.assertFalse(repo.is_locked())
        self.assertFalse(fallback_repo.is_locked())



# Copyright (C) 2007, 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import errno
import smtplib
import socket
from email.message import Message

from breezy import config, email_message, smtp_connection, tests, ui


def connection_refuser(host):
    raise OSError(errno.ECONNREFUSED, 'Connection Refused')


class StubSMTPFactory:
    """A fake SMTP connection to test the connection setup."""

    def __init__(self, fail_on=None, smtp_features=None):
        self._fail_on = fail_on or []
        self._calls = []
        self._smtp_features = smtp_features or []
        self._ehlo_called = False

    def __call__(self, host='localhost'):
        self._calls.append(('connect', host))
        # The factory pretends to be a connection
        return self

    def connect(self, server):
        raise NotImplementedError

    def helo(self):
        self._calls.append(('helo',))
        if 'helo' in self._fail_on:
            return 500, 'helo failure'
        else:
            return 200, 'helo success'

    def ehlo(self):
        self._calls.append(('ehlo',))
        if 'ehlo' in self._fail_on:
            return 500, 'ehlo failure'
        else:
            self._ehlo_called = True
            return 200, 'ehlo success'

    def has_extn(self, extension):
        self._calls.append(('has_extn', extension))
        return self._ehlo_called and extension in self._smtp_features

    def starttls(self):
        self._calls.append(('starttls',))
        if 'starttls' in self._fail_on:
            return 500, 'starttls failure'
        else:
            self._ehlo_called = True
            return 200, 'starttls success'


class WideOpenSMTPFactory(StubSMTPFactory):
    """A fake smtp server that implements login by accepting anybody."""

    def login(self, user, password):
        self._calls.append(('login', user, password))


class TestSMTPConnection(tests.TestCaseInTempDir):

    def get_connection(self, text, smtp_factory=None):
        my_config = config.MemoryStack(text)
        return smtp_connection.SMTPConnection(
            my_config, _smtp_factory=smtp_factory)

    def test_defaults(self):
        conn = self.get_connection(b'')
        self.assertEqual('localhost', conn._smtp_server)
        self.assertEqual(None, conn._smtp_username)
        self.assertEqual(None, conn._smtp_password)

    def test_smtp_server(self):
        conn = self.get_connection(b'smtp_server=host:10')
        self.assertEqual('host:10', conn._smtp_server)

    def test_missing_server(self):
        conn = self.get_connection(b'', smtp_factory=connection_refuser)
        self.assertRaises(smtp_connection.DefaultSMTPConnectionRefused,
                          conn._connect)
        conn = self.get_connection(b'smtp_server=smtp.example.com',
                                   smtp_factory=connection_refuser)
        self.assertRaises(smtp_connection.SMTPConnectionRefused, conn._connect)

    def test_smtp_username(self):
        conn = self.get_connection(b'')
        self.assertIs(None, conn._smtp_username)

        conn = self.get_connection(b'smtp_username=joebody')
        self.assertEqual('joebody', conn._smtp_username)

    def test_smtp_password_from_config(self):
        conn = self.get_connection(b'')
        self.assertIs(None, conn._smtp_password)

        conn = self.get_connection(b'smtp_password=mypass')
        self.assertEqual('mypass', conn._smtp_password)

    def test_smtp_password_from_user(self):
        user = 'joe'
        password = 'hispass'
        factory = WideOpenSMTPFactory()
        conn = self.get_connection(
            b'[DEFAULT]\nsmtp_username=%s\n' % user.encode('ascii'),
            smtp_factory=factory)
        self.assertIs(None, conn._smtp_password)

        ui.ui_factory = ui.CannedInputUIFactory([password])
        conn._connect()
        self.assertEqual(password, conn._smtp_password)

    def test_smtp_password_from_auth_config(self):
        user = 'joe'
        password = 'hispass'
        factory = WideOpenSMTPFactory()
        conn = self.get_connection(
            b'[DEFAULT]\nsmtp_username=%s\n' % user.encode('ascii'),
            smtp_factory=factory)
        self.assertEqual(user, conn._smtp_username)
        self.assertIs(None, conn._smtp_password)
        # Create a config file with the right password
        conf = config.AuthenticationConfig()
        conf._get_config().update({'smtptest':
                                   {'scheme': 'smtp', 'user': user,
                                    'password': password}})
        conf._save()

        conn._connect()
        self.assertEqual(password, conn._smtp_password)

    def test_authenticate_with_byte_strings(self):
        user = b'joe'
        unicode_pass = 'h\xECspass'
        utf8_pass = unicode_pass.encode('utf-8')
        factory = WideOpenSMTPFactory()
        conn = self.get_connection(
            b'[DEFAULT]\nsmtp_username=%s\nsmtp_password=%s\n'
            % (user, utf8_pass), smtp_factory=factory)
        self.assertEqual(unicode_pass, conn._smtp_password)
        conn._connect()
        self.assertEqual([('connect', 'localhost'),
                          ('ehlo',),
                          ('has_extn', 'starttls'),
                          ('login', user, utf8_pass)], factory._calls)
        smtp_username, smtp_password = factory._calls[-1][1:]
        self.assertIsInstance(smtp_username, bytes)
        self.assertIsInstance(smtp_password, bytes)

    def test_create_connection(self):
        factory = StubSMTPFactory()
        conn = self.get_connection(b'', smtp_factory=factory)
        conn._create_connection()
        self.assertEqual([('connect', 'localhost'),
                          ('ehlo',),
                          ('has_extn', 'starttls')], factory._calls)

    def test_create_connection_ehlo_fails(self):
        # Check that we call HELO if EHLO failed.
        factory = StubSMTPFactory(fail_on=['ehlo'])
        conn = self.get_connection(b'', smtp_factory=factory)
        conn._create_connection()
        self.assertEqual([('connect', 'localhost'),
                          ('ehlo',),
                          ('helo',),
                          ('has_extn', 'starttls')], factory._calls)

    def test_create_connection_ehlo_helo_fails(self):
        # Check that we raise an exception if both EHLO and HELO fail.
        factory = StubSMTPFactory(fail_on=['ehlo', 'helo'])
        conn = self.get_connection(b'', smtp_factory=factory)
        self.assertRaises(smtp_connection.SMTPError, conn._create_connection)
        self.assertEqual([('connect', 'localhost'),
                          ('ehlo',),
                          ('helo',)], factory._calls)

    def test_create_connection_starttls(self):
        # Check that STARTTLS plus a second EHLO are called if the
        # server says it supports the feature.
        factory = StubSMTPFactory(smtp_features=['starttls'])
        conn = self.get_connection(b'', smtp_factory=factory)
        conn._create_connection()
        self.assertEqual([('connect', 'localhost'),
                          ('ehlo',),
                          ('has_extn', 'starttls'),
                          ('starttls',),
                          ('ehlo',)], factory._calls)

    def test_create_connection_starttls_fails(self):
        # Check that we raise an exception if the server claims to
        # support STARTTLS, but then fails when we try to activate it.
        factory = StubSMTPFactory(
            fail_on=['starttls'],
            smtp_features=['starttls'])
        conn = self.get_connection(b'', smtp_factory=factory)
        self.assertRaises(smtp_connection.SMTPError, conn._create_connection)
        self.assertEqual([('connect', 'localhost'),
                          ('ehlo',),
                          ('has_extn', 'starttls'),
                          ('starttls',)], factory._calls)

    def test_get_message_addresses(self):
        msg = Message()

        from_, to = smtp_connection.SMTPConnection.get_message_addresses(msg)
        self.assertEqual('', from_)
        self.assertEqual([], to)

        msg['From'] = '"J. Random Developer" <jrandom@example.com>'
        msg['To'] = 'John Doe <john@doe.com>, Jane Doe <jane@doe.com>'
        msg['CC'] = 'Pepe P\xe9rez <pperez@ejemplo.com>'
        msg['Bcc'] = 'user@localhost'

        from_, to = smtp_connection.SMTPConnection.get_message_addresses(msg)
        self.assertEqual('jrandom@example.com', from_)
        self.assertEqual(sorted(['john@doe.com', 'jane@doe.com',
                                 'pperez@ejemplo.com', 'user@localhost']), sorted(to))

        # now with breezy's EmailMessage
        msg = email_message.EmailMessage(
            '"J. Random Developer" <jrandom@example.com>',
            ['John Doe <john@doe.com>', 'Jane Doe <jane@doe.com>',
             'Pepe P\xe9rez <pperez@ejemplo.com>', 'user@localhost'],
            'subject')

        from_, to = smtp_connection.SMTPConnection.get_message_addresses(msg)
        self.assertEqual('jrandom@example.com', from_)
        self.assertEqual(sorted(['john@doe.com', 'jane@doe.com',
                                 'pperez@ejemplo.com', 'user@localhost']), sorted(to))

    def test_destination_address_required(self):
        msg = Message()
        msg['From'] = '"J. Random Developer" <jrandom@example.com>'
        self.assertRaises(
            smtp_connection.NoDestinationAddress,
            smtp_connection.SMTPConnection(config.MemoryStack(b"")
                                           ).send_email, msg)

        msg = email_message.EmailMessage('from@from.com', '', 'subject')
        self.assertRaises(
            smtp_connection.NoDestinationAddress,
            smtp_connection.SMTPConnection(config.MemoryStack(b"")
                                           ).send_email, msg)

        msg = email_message.EmailMessage('from@from.com', [], 'subject')
        self.assertRaises(
            smtp_connection.NoDestinationAddress,
            smtp_connection.SMTPConnection(config.MemoryStack(b"")
                                           ).send_email, msg)



# Copyright (C) 2006, 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Adapter for running test cases against multiple encodings."""

# prefix for micro (1/1000000)
_mu = '\xb5'

# greek letter omega, not to be confused with
# the Ohm sign, u'\u2126'. Though they are probably identical
# cp437 can handle the first, but not the second
_omega = '\u03a9'

# smallest error possible, epsilon
# cp437 handles u03b5, but not u2208 the 'element of' operator
_epsilon = '\u03b5'

# Swedish?
_erik = 'Erik B\xe5gfors'

# Swedish 'räksmörgås' means shrimp sandwich
_shrimp_sandwich = 'r\xe4ksm\xf6rg\xe5s'

# Arabic, probably only Unicode encodings can handle this one
_juju = '\u062c\u0648\u062c\u0648'

# iso-8859-1 alternative for juju
_juju_alt = 'j\xfbj\xfa'

# Russian, 'Alexander' in russian
_alexander = '\u0410\u043b\u0435\u043a\u0441\u0430\u043d\u0434\u0440'
# The word 'test' in Russian
_russian_test = '\u0422\u0435\u0441\u0442'

# Kanji
# It is a kanji sequence for nihonjin, or Japanese in English.
#
# '\u4eba' being person, 'u\65e5' sun and '\u672c' origin. Ie,
# sun-origin-person, 'native from the land where the sun rises'. Note, I'm
# not a fluent speaker, so this is just my crude breakdown.
#
# Wouter van Heyst
_nihonjin = '\u65e5\u672c\u4eba'

# Czech
# It's what is usually used for showing how fonts look, because it contains
# most accented characters, ie. in places where Englishman use 'Quick brown fox
# jumped over a lazy dog'. The literal translation of the Czech version would
# be something like 'Yellow horse groaned devilish codes'. Actually originally
# the last word used to be 'ódy' (odes). The 'k' was added as a pun when using
# the sentece to check whether one has properly set encoding.
_yellow_horse = ('\u017dlu\u0165ou\u010dk\xfd k\u016f\u0148'
                 ' \xfap\u011bl \u010f\xe1belsk\xe9 k\xf3dy')
_yellow = '\u017dlu\u0165ou\u010dk\xfd'
_someone = 'Some\u016f\u0148\u011b'
_something = '\u0165ou\u010dk\xfd'

# Hebrew
# Shalom -> 'hello' or 'peace', used as a common greeting
_shalom = '\u05e9\u05dc\u05d5\u05dd'


encoding_scenarios = [
    # Permutation 1 of utf-8
    ('utf-8,1', {
        'info': {
            'committer': _erik,
            'message': _yellow_horse,
            'filename': _shrimp_sandwich,
            'directory': _nihonjin,
            },
        'encoding': 'utf-8',
        }),
    # Permutation 2 of utf-8
    ('utf-8,2', {
        'info': {
            'committer': _alexander,
            'message': 'Testing ' + _mu,
            'filename': _shalom,
            'directory': _juju,
            },
        'encoding': 'utf-8',
        }),
    ('iso-8859-1', {
        'info': {
            'committer': _erik,
            'message': 'Testing ' + _mu,
            'filename': _juju_alt,
            'directory': _shrimp_sandwich,
            },
        'encoding': 'iso-8859-1',
        }),
    ('iso-8859-2', {
        'info': {
            'committer': _someone,
            'message': _yellow_horse,
            'filename': _yellow,
            'directory': _something,
            },
        'encoding': 'iso-8859-2',
        }),
    ('cp1251', {
        'info': {
            'committer': _alexander,
            'message': 'Testing ' + _mu,
            'filename': _russian_test,
            'directory': _russian_test + 'dir',
            },
        'encoding': 'cp1251',
        }),
    # The iso-8859-1 tests run on a default windows cp437 installation
    # and it takes a long time to run an extra permutation of the tests
    # But just in case we want to add this back in:
    #        ('cp437', {'committer':_erik
    #                  , 'message':u'Testing ' + _mu
    #                  , 'filename':'file_' + _omega
    #                  , 'directory':_epsilon + '_dir',
    #            'encoding': 'cp437'}),
    ]



# Copyright (C) 2010 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License or
# (at your option) a later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Darcs foreign branch support.

Currently only tells the user to use fastimport/fastexport.
"""

from breezy import controldir, errors

from ... import version_info  # noqa: F401


class DarcsUnsupportedError(errors.UnsupportedVcs):

    vcs = "darcs"

    _fmt = ('Darcs branches are not yet supported. '
            'To interoperate with darcs branches, use fastimport.')


class DarcsDirFormat(controldir.ControlDirFormat):
    """Darcs directory format."""

    def get_converter(self):
        raise NotImplementedError(self.get_converter)

    def get_format_description(self):
        return "darcs control directory"

    def initialize_on_transport(self, transport):
        raise errors.UninitializableFormat(self)

    def is_supported(self):
        return False

    def supports_transport(self, transport):
        return False

    @classmethod
    def _known_formats(self):
        return {DarcsDirFormat()}

    def check_support_status(self, allow_unsupported, recommend_upgrade=True,
                             basedir=None):
        raise DarcsUnsupportedError()

    def open(self, transport):
        # Raise NotBranchError if there is nothing there
        DarcsProber().probe_transport(transport)
        raise NotImplementedError(self.open)


class DarcsProber(controldir.Prober):

    @classmethod
    def priority(klass, transport):
        if 'darcs' in transport.base:
            return 90
        return 100

    @classmethod
    def probe_transport(klass, transport):
        if transport.has_any(['_darcs/format', '_darcs/inventory']):
            return DarcsDirFormat()
        raise errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        return [DarcsDirFormat()]


controldir.ControlDirFormat.register_prober(DarcsProber)



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Interface thunk for a StaticTuple implementation."""

from .. import debug

try:
    from ._static_tuple_c import StaticTuple
except ImportError as e:
    from .. import osutils
    osutils.failed_to_load_extension(e)
    from ._static_tuple_py import StaticTuple


def expect_static_tuple(obj):
    """Check if the passed object is a StaticTuple.

    Cast it if necessary, but if the 'static_tuple' debug flag is set, raise an
    error instead.

    As apis are improved, we will probably eventually stop calling this as it
    adds overhead we shouldn't need.
    """
    if 'static_tuple' not in debug.debug_flags:
        return StaticTuple.from_sequence(obj)
    if not isinstance(obj, StaticTuple):
        raise TypeError('We expected a StaticTuple not a {}'.format(type(obj)))
    return obj


def as_tuples(obj):
    """Ensure that the object and any referenced objects are plain tuples.

    :param obj: a list, tuple or StaticTuple
    :return: a plain tuple instance, with all children also being tuples.
    """
    result = []
    for item in obj:
        if isinstance(item, (tuple, list, StaticTuple)):
            item = as_tuples(item)
        result.append(item)
    return tuple(result)



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Branch.iter_merge_sorted_revisions()"""

from breezy import errors, revision, tests
from breezy.tests import per_branch


class TestIterMergeSortedRevisionsSimpleGraph(per_branch.TestCaseWithBranch):

    def setUp(self):
        super().setUp()
        self.revids = {}
        builder = self.make_builder_with_merges('.')
        self.branch = builder.get_branch()
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)

    def make_snapshot(self, builder, parents, revid_name):
        self.assertNotIn(revid_name, self.revids)
        if parents is None:
            files = [('add', ('', None, 'directory', '')), ]
        else:
            parents = [self.revids[name] for name in parents]
            files = []
        self.revids[revid_name] = builder.build_snapshot(
            parents, files, message="Revision %s" % revid_name)

    def make_builder_with_merges(self, relpath):
        try:
            builder = self.make_branch_builder(relpath)
        except (errors.TransportNotPossible, errors.UninitializableFormat):
            raise tests.TestNotApplicable('format not directly constructable')
        builder.start_series()
        # 1
        # |\
        # 2 |
        # | |
        # | 1.1.1
        # |/
        # 3
        self.make_snapshot(builder, None, '1')
        self.make_snapshot(builder, ['1'], '1.1.1')
        self.make_snapshot(builder, ['1'], '2')
        self.make_snapshot(builder, ['2', '1.1.1'], '3')
        builder.finish_series()
        return builder

    def assertIterRevids(self, expected, *args, **kwargs):
        if kwargs.get('stop_revision_id') is not None:
            kwargs['stop_revision_id'] = self.revids[kwargs['stop_revision_id']]
        if kwargs.get('start_revision_id') is not None:
            kwargs['start_revision_id'] = self.revids[kwargs['start_revision_id']]
        # We don't care about depths and revnos here, only about returning the
        # right revids.
        revids = [revid for (revid, depth, revno, eom) in
                  self.branch.iter_merge_sorted_revisions(*args, **kwargs)]
        self.assertEqual([self.revids[short] for short in expected], revids)

    def test_merge_sorted(self):
        self.assertIterRevids(['3', '1.1.1', '2', '1'])

    def test_merge_sorted_range(self):
        self.assertIterRevids(['1.1.1'],
                              start_revision_id='1.1.1', stop_revision_id='1')

    def test_merge_sorted_range_start_only(self):
        self.assertIterRevids(['1.1.1', '1'],
                              start_revision_id='1.1.1')

    def test_merge_sorted_range_stop_exclude(self):
        self.assertIterRevids(['3', '1.1.1', '2'], stop_revision_id='1')

    def test_merge_sorted_range_stop_include(self):
        self.assertIterRevids(['3', '1.1.1', '2'],
                              stop_revision_id='2', stop_rule='include')

    def test_merge_sorted_range_stop_with_merges(self):
        self.assertIterRevids(['3', '1.1.1'],
                              stop_revision_id='3', stop_rule='with-merges')

    def test_merge_sorted_range_stop_with_merges_can_show_non_parents(self):
        # 1.1.1 gets logged before the end revision is reached.
        # so it is returned even though 1.1.1 is not a parent of 2.
        self.assertIterRevids(['3', '1.1.1', '2'],
                              stop_revision_id='2', stop_rule='with-merges')

    def test_merge_sorted_range_stop_with_merges_ignore_non_parents(self):
        # 2 is not a parent of 1.1.1 so it must not be returned
        self.assertIterRevids(['3', '1.1.1'],
                              stop_revision_id='1.1.1', stop_rule='with-merges')

    def test_merge_sorted_single_stop_exclude(self):
        # from X..X exclusive is an empty result
        self.assertIterRevids([], start_revision_id='3', stop_revision_id='3')

    def test_merge_sorted_single_stop_include(self):
        # from X..X inclusive is [X]
        self.assertIterRevids(['3'],
                              start_revision_id='3', stop_revision_id='3',
                              stop_rule='include')

    def test_merge_sorted_single_stop_with_merges(self):
        self.assertIterRevids(['3', '1.1.1'],
                              start_revision_id='3', stop_revision_id='3',
                              stop_rule='with-merges')

    def test_merge_sorted_forward(self):
        self.assertIterRevids(['1', '2', '1.1.1', '3'], direction='forward')

    def test_merge_sorted_range_forward(self):
        self.assertIterRevids(['1.1.1'],
                              start_revision_id='1.1.1', stop_revision_id='1',
                              direction='forward')

    def test_merge_sorted_range_start_only_forward(self):
        self.assertIterRevids(['1', '1.1.1'],
                              start_revision_id='1.1.1', direction='forward')

    def test_merge_sorted_range_stop_exclude_forward(self):
        self.assertIterRevids(['2', '1.1.1', '3'],
                              stop_revision_id='1', direction='forward')

    def test_merge_sorted_range_stop_include_forward(self):
        self.assertIterRevids(['2', '1.1.1', '3'],
                              stop_revision_id='2', stop_rule='include',
                              direction='forward')

    def test_merge_sorted_range_stop_with_merges_forward(self):
        self.assertIterRevids(['1.1.1', '3'],
                              stop_revision_id='3', stop_rule='with-merges',
                              direction='forward')


class TestIterMergeSortedRevisionsBushyGraph(per_branch.TestCaseWithBranch):

    def setUp(self):
        super().setUp()
        self.revids = {}

    def make_branch_builder(self, relpath):
        try:
            builder = super().make_branch_builder(relpath)
        except (errors.TransportNotPossible, errors.UninitializableFormat):
            raise tests.TestNotApplicable('format not directly constructable')
        return builder

    def make_snapshot(self, builder, parents, revid_name):
        self.assertNotIn(revid_name, self.revids)
        if parents is None:
            files = [('add', ('', None, 'directory', '')), ]
        else:
            parents = [self.revids[name] for name in parents]
            files = []
        self.revids[revid_name] = builder.build_snapshot(
            parents, files, message="Revision %s" % revid_name)

    def make_branch_with_embedded_merges(self, relpath='.'):
        builder = self.make_branch_builder(relpath)

        # 1
        # |\
        # | 1.1.1
        # | /
        # 2
        # | \
        # 3 |
        # | |
        # | 2.1.1
        # | |    \
        # | 2.1.2 |
        # | |     |
        # | |   2.2.1
        # | |  /
        # | 2.1.3
        # |/
        # 4
        builder.start_series()
        self.make_snapshot(builder, None, '1')
        self.make_snapshot(builder, ['1'], '1.1.1')
        self.make_snapshot(builder, ['1', '1.1.1'], '2')
        self.make_snapshot(builder, ['2'], '2.1.1')
        self.make_snapshot(builder, ['2.1.1'], '2.1.2')
        self.make_snapshot(builder, ['2.1.1'], '2.2.1')
        self.make_snapshot(builder, ['2.1.2', '2.2.1'], '2.1.3')
        self.make_snapshot(builder, ['2'], '3')
        self.make_snapshot(builder, ['3', '2.1.3'], '4')
        builder.finish_series()
        br = builder.get_branch()
        br.lock_read()
        self.addCleanup(br.unlock)
        return br

    def make_branch_with_different_depths_merges(self, relpath='.'):
        builder = self.make_branch_builder(relpath)
        # 1
        # |\
        # | 1.1.1
        # | /
        # 2
        # | \
        # 3 |
        # | |
        # | 2.1.1
        # | |    \
        # | 2.1.2 |
        # | |     |
        # | |     2.2.1
        # | |    /
        # | 2.1.3
        # |/
        # 4
        builder.start_series()
        self.make_snapshot(builder, None, '1')
        self.make_snapshot(builder, ['1'], '2')
        self.make_snapshot(builder, ['1'], '1.1.1')
        self.make_snapshot(builder, ['1.1.1'], '1.1.2')
        self.make_snapshot(builder, ['1.1.1'], '1.2.1')
        self.make_snapshot(builder, ['1.2.1'], '1.2.2')
        self.make_snapshot(builder, ['1.2.1'], '1.3.1')
        self.make_snapshot(builder, ['1.3.1'], '1.3.2')
        self.make_snapshot(builder, ['1.3.1'], '1.4.1')
        self.make_snapshot(builder, ['1.3.2'], '1.3.3')
        self.make_snapshot(builder, ['1.2.2', '1.3.3'], '1.2.3')
        self.make_snapshot(builder, ['2'], '2.1.1')
        self.make_snapshot(builder, ['2.1.1'], '2.1.2')
        self.make_snapshot(builder, ['2.1.1'], '2.2.1')
        self.make_snapshot(builder, ['2.1.2', '2.2.1'], '2.1.3')
        self.make_snapshot(builder, ['2', '1.2.3'], '3')
        # .. to bring them all and ... bind them
        self.make_snapshot(builder, ['3', '2.1.3'], '4')
        builder.finish_series()
        br = builder.get_branch()
        br.lock_read()
        self.addCleanup(br.unlock)
        return br

    def make_branch_with_alternate_ancestries(self, relpath='.'):
        # See test_merge_sorted_exclude_ancestry below for the difference with
        # bt.test_log.TestLogExcludeAncestry.
        # make_branch_with_alternate_ancestries and
        # test_merge_sorted_exclude_ancestry
        # See the FIXME in assertLogRevnos there too.
        builder = self.make_branch_builder(relpath)
        # 1
        # |\
        # | 1.1.1
        # | /| \
        # 2  |  |
        # |  |  1.2.1
        # |  | /
        # |  1.1.2
        # | /
        # 3
        builder.start_series()
        self.make_snapshot(builder, None, '1')
        self.make_snapshot(builder, ['1'], '1.1.1')
        self.make_snapshot(builder, ['1', '1.1.1'], '2')
        self.make_snapshot(builder, ['1.1.1'], '1.2.1')
        self.make_snapshot(builder, ['1.1.1', '1.2.1'], '1.1.2')
        self.make_snapshot(builder, ['2', '1.1.2'], '3')
        builder.finish_series()
        br = builder.get_branch()
        br.lock_read()
        self.addCleanup(br.unlock)
        return br

    def assertIterRevids(self, expected, branch, *args, **kwargs):
        if kwargs.get('stop_revision_id') is not None:
            kwargs['stop_revision_id'] = self.revids[kwargs['stop_revision_id']]
        if kwargs.get('start_revision_id') is not None:
            kwargs['start_revision_id'] = self.revids[kwargs['start_revision_id']]
        # We don't care about depths and revnos here, only about returning the
        # right revids.
        revids = [revid for (revid, depth, revno, eom) in
                  branch.iter_merge_sorted_revisions(*args, **kwargs)]
        self.assertEqual([self.revids[short] for short in expected], revids)

    def test_merge_sorted_starting_at_embedded_merge(self):
        branch = self.make_branch_with_embedded_merges()
        self.assertIterRevids(['4', '2.1.3', '2.2.1', '2.1.2', '2.1.1',
                               '3', '2', '1.1.1', '1'],
                              branch)
        # 3 and 2.1.2 are not part of 2.2.1 ancestry and should not appear
        self.assertIterRevids(['2.2.1', '2.1.1', '2', '1.1.1', '1'],
                              branch, start_revision_id='2.2.1',
                              stop_rule='with-merges')

    def test_merge_sorted_with_different_depths_merge(self):
        branch = self.make_branch_with_different_depths_merges()
        self.assertIterRevids(['4', '2.1.3', '2.2.1', '2.1.2', '2.1.1',
                               '3',
                               '1.2.3', '1.3.3', '1.3.2', '1.3.1',
                               '1.2.2', '1.2.1', '1.1.1',
                               '2', '1'],
                              branch)
        # 3 (and its descendants) and 2.1.2 are not part of 2.2.1 ancestry and
        # should not appear
        self.assertIterRevids(['2.2.1', '2.1.1', '2', '1'],
                              branch, start_revision_id='2.2.1',
                              stop_rule='with-merges')

    def test_merge_sorted_exclude_ancestry(self):
        branch = self.make_branch_with_alternate_ancestries()
        self.assertIterRevids(['3', '1.1.2', '1.2.1', '2', '1.1.1', '1'],
                              branch)
        # '2' is not part of the ancestry even if merge_sort order will make it
        # appear before 1.1.1
        self.assertIterRevids(['1.1.2', '1.2.1'],
                              branch,
                              stop_rule='with-merges-without-common-ancestry',
                              start_revision_id='1.1.2',
                              stop_revision_id='1.1.1')



# Copyright (C) 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import errors, tests
from breezy.tests.per_tree import TestCaseWithTree


class Path2IdTests(TestCaseWithTree):

    def setUp(self):
        super().setUp()
        work_a = self.make_branch_and_tree('wta')
        if not work_a.supports_setting_file_ids():
            self.skipTest("working tree does not support setting file ids")
        self.build_tree(['wta/bla', 'wta/dir/', 'wta/dir/file'])
        work_a.add(['bla', 'dir', 'dir/file'],
                   ids=[b'bla-id', b'dir-id', b'file-id'])
        work_a.commit('add files')
        self.tree_a = self.workingtree_to_test_tree(work_a)

    def test_path2id(self):
        self.assertEqual(b'bla-id', self.tree_a.path2id('bla'))
        self.assertEqual(b'dir-id', self.tree_a.path2id('dir'))
        self.assertIs(None, self.tree_a.path2id('idontexist'))

    def test_path2id_list(self):
        self.assertEqual(b'bla-id', self.tree_a.path2id(['bla']))
        self.assertEqual(b'dir-id', self.tree_a.path2id(['dir']))
        self.assertEqual(b'file-id', self.tree_a.path2id(['dir', 'file']))
        self.assertEqual(self.tree_a.path2id(''),
                         self.tree_a.path2id([]))
        self.assertIs(None, self.tree_a.path2id(['idontexist']))
        self.assertIs(None, self.tree_a.path2id(['dir', 'idontexist']))

    def test_id2path(self):
        self.addCleanup(self.tree_a.lock_read().unlock)
        self.assertEqual('bla', self.tree_a.id2path(b'bla-id'))
        self.assertEqual('dir', self.tree_a.id2path(b'dir-id'))
        self.assertEqual('dir/file', self.tree_a.id2path(b'file-id'))
        self.assertRaises(errors.NoSuchId, self.tree_a.id2path, b'nonexistant')

    def skip_if_no_reference(self, tree):
        if not getattr(tree, 'supports_tree_reference', lambda: False)():
            raise tests.TestNotApplicable('Tree references not supported')

    def create_nested(self):
        work_tree = self.make_branch_and_tree('wt')
        with work_tree.lock_write():
            self.skip_if_no_reference(work_tree)
            subtree = self.make_branch_and_tree('wt/subtree')
            self.build_tree(['wt/subtree/a'])
            subtree.add(['a'])
            subtree.commit('foo')
            work_tree.add_reference(subtree)
        tree = self._convert_tree(work_tree)
        self.skip_if_no_reference(tree)
        return tree, subtree

    def test_path2id_nested_tree(self):
        tree, subtree = self.create_nested()
        self.assertIsNot(None, tree.path2id('subtree'))
        self.assertIsNot(None, tree.path2id('subtree/a'))
        self.assertEqual('subtree', tree.id2path(tree.path2id('subtree')))
        self.assertEqual('subtree/a', tree.id2path(tree.path2id('subtree/a')))
        self.assertIsNot('subtree/a', tree.id2path(tree.path2id('subtree/a'), recurse='down'))
        self.assertRaises(errors.NoSuchId, tree.id2path, tree.path2id('subtree/a'), recurse='none')


class Path2IdsTests(TestCaseWithTree):

    def test_paths2ids_recursive(self):
        work_tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/dir/', 'tree/dir/file'])
        work_tree.add(['dir', 'dir/file'])
        if not work_tree.supports_setting_file_ids():
            raise tests.TestNotApplicable(
                "test not applicable on non-inventory tests")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual({tree.path2id('dir'), tree.path2id('dir/file')},
                         tree.paths2ids(['dir']))

    def test_paths2ids_forget_old(self):
        work_tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/file'])
        work_tree.add('file')
        work_tree.commit('commit old state')
        work_tree.remove('file')
        if not work_tree.supports_setting_file_ids():
            raise tests.TestNotApplicable(
                "test not applicable on non-inventory tests")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(set(), tree.paths2ids(['file'],
                                                 require_versioned=False))



# Copyright (C) 2005 Canonical Ltd

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"bash_completion.py - create bash completion script from built-in brz help"

import breezy
import breezy.commands
import breezy.help
from breezy.doc_generate import get_autodoc_datetime


def get_filename(options):
    return "%s.bash_completion" % (options.brz_name)


def infogen(options, outfile):
    d = get_autodoc_datetime()
    params = \
        {"brzcmd": options.brz_name,
         "datestamp": d.strftime("%Y-%m-%d"),
         "timestamp": d.strftime("%Y-%m-%d %H:%M:%S +0000"),
         "version": breezy.__version__,
         }

    outfile.write(preamble % params)


preamble = """\
# bash completion functions for for Breezy (%(brzcmd)s)
#
# Large parts of this file are autogenerated from the internal
# Breezy documentation and data structures.
#
# Generation time: %(timestamp)s
"""



# Copyright (C) 2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__doc__ = """Generate a shell function for zsh command line completion.
"""

from ... import commands, version_info  # noqa: F401

bzr_plugin_name = 'zsh_completion'
bzr_commands = ['zsh-completion']

commands.plugin_cmds.register_lazy('cmd_zsh_completion', [],
                                   __name__ + '.zshcomp')


def load_tests(loader, basic_tests, pattern):
    testmod_names = [
        'tests',
        ]
    basic_tests.addTest(loader.loadTestsFromModuleNames(
        ["{}.{}".format(__name__, tmn) for tmn in testmod_names]))
    return basic_tests



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""B+Tree index parsing."""

from . import static_tuple


def _parse_leaf_lines(data, key_length, ref_list_length):
    lines = data.split(b'\n')
    nodes = []
    as_st = static_tuple.StaticTuple.from_sequence
    stuple = static_tuple.StaticTuple
    for line in lines[1:]:
        if line == b'':
            return nodes
        elements = line.split(b'\0', key_length)
        # keys are tuples
        key = as_st(elements[:key_length]).intern()
        line = elements[-1]
        references, value = line.rsplit(b'\0', 1)
        if ref_list_length:
            ref_lists = []
            for ref_string in references.split(b'\t'):
                ref_list = as_st([as_st(ref.split(b'\0')).intern()
                                  for ref in ref_string.split(b'\r') if ref])
                ref_lists.append(ref_list)
            ref_lists = as_st(ref_lists)
            node_value = stuple(value, ref_lists)
        else:
            node_value = stuple(value, stuple())
        # No need for StaticTuple here as it is put into a dict
        nodes.append((key, node_value))
    return nodes


def _flatten_node(node, reference_lists):
    """Convert a node into the serialized form.

    :param node: A tuple representing a node (key_tuple, value, references)
    :param reference_lists: Does this index have reference lists?
    :return: (string_key, flattened)
        string_key  The serialized key for referencing this node
        flattened   A string with the serialized form for the contents
    """
    if reference_lists:
        # TODO: Consider turning this back into the 'unoptimized' nested loop
        #       form. It is probably more obvious for most people, and this is
        #       just a reference implementation.
        flattened_references = [b'\r'.join([b'\x00'.join(reference)
                                            for reference in ref_list])
                                for ref_list in node[3]]
    else:
        flattened_references = []
    string_key = b'\x00'.join(node[1])
    line = (b"%s\x00%s\x00%s\n" % (string_key,
                                   b'\t'.join(flattened_references), node[2]))
    return string_key, line



# Copyright (C) 2005-2011, 2015, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import errno
import os
import subprocess
import sys
import threading
from io import BytesIO

import breezy.transport.trace

from .. import errors, osutils, tests, transport, urlutils
from ..transport import (FileExists, NoSuchFile, UnsupportedProtocol, chroot,
                         fakenfs, http, local, memory, pathfilter, readonly)
from . import features, test_server

# TODO: Should possibly split transport-specific tests into their own files.


class TestTransport(tests.TestCase):
    """Test the non transport-concrete class functionality."""

    def test__get_set_protocol_handlers(self):
        handlers = transport._get_protocol_handlers()
        self.assertNotEqual([], handlers.keys())
        transport._clear_protocol_handlers()
        self.addCleanup(transport._set_protocol_handlers, handlers)
        self.assertEqual([], transport._get_protocol_handlers().keys())

    def test_get_transport_modules(self):
        handlers = transport._get_protocol_handlers()
        self.addCleanup(transport._set_protocol_handlers, handlers)
        # don't pollute the current handlers
        transport._clear_protocol_handlers()

        class SampleHandler:
            """I exist, isnt that enough?"""
        transport._clear_protocol_handlers()
        transport.register_transport_proto('foo')
        transport.register_lazy_transport('foo',
                                          'breezy.tests.test_transport',
                                          'TestTransport.SampleHandler')
        transport.register_transport_proto('bar')
        transport.register_lazy_transport('bar',
                                          'breezy.tests.test_transport',
                                          'TestTransport.SampleHandler')
        self.assertEqual([SampleHandler.__module__,
                          'breezy.transport.chroot',
                          'breezy.transport.pathfilter'],
                         transport._get_transport_modules())

    def test_transport_dependency(self):
        """Transport with missing dependency causes no error"""
        saved_handlers = transport._get_protocol_handlers()
        self.addCleanup(transport._set_protocol_handlers, saved_handlers)
        # don't pollute the current handlers
        transport._clear_protocol_handlers()
        transport.register_transport_proto('foo')
        transport.register_lazy_transport(
            'foo', 'breezy.tests.test_transport', 'BadTransportHandler')
        try:
            transport.get_transport_from_url('foo://fooserver/foo')
        except UnsupportedProtocol as e:
            self.assertEqual('Unsupported protocol'
                             ' for url "foo://fooserver/foo":'
                             ' Unable to import library "some_lib":'
                             ' testing missing dependency', str(e))
        else:
            self.fail('Did not raise UnsupportedProtocol')

    def test_transport_fallback(self):
        """Transport with missing dependency causes no error"""
        saved_handlers = transport._get_protocol_handlers()
        self.addCleanup(transport._set_protocol_handlers, saved_handlers)
        transport._clear_protocol_handlers()
        transport.register_transport_proto('foo')
        transport.register_lazy_transport(
            'foo', 'breezy.tests.test_transport', 'BackupTransportHandler')
        transport.register_lazy_transport(
            'foo', 'breezy.tests.test_transport', 'BadTransportHandler')
        t = transport.get_transport_from_url('foo://fooserver/foo')
        self.assertTrue(isinstance(t, BackupTransportHandler))

    def test_ssh_hints(self):
        """Transport ssh:// should raise an error pointing out bzr+ssh://"""
        try:
            transport.get_transport_from_url('ssh://fooserver/foo')
        except UnsupportedProtocol as e:
            self.assertEqual(
                'Unsupported protocol'
                ' for url "ssh://fooserver/foo":'
                ' Use bzr+ssh for Bazaar operations over SSH, '
                'e.g. "bzr+ssh://fooserver/foo". Use git+ssh '
                'for Git operations over SSH, e.g. "git+ssh://fooserver/foo".',
                str(e))
        else:
            self.fail('Did not raise UnsupportedProtocol')

    def test_LateReadError(self):
        """The LateReadError helper should raise on read()."""
        a_file = transport.LateReadError('a path')
        try:
            a_file.read()
        except errors.ReadError as error:
            self.assertEqual('a path', error.path)
        self.assertRaises(errors.ReadError, a_file.read, 40)
        a_file.close()

    def test_local_abspath_non_local_transport(self):
        # the base implementation should throw
        t = memory.MemoryTransport()
        e = self.assertRaises(errors.NotLocalUrl, t.local_abspath, 't')
        self.assertEqual('memory:///t is not a local path.', str(e))


class TestCoalesceOffsets(tests.TestCase):

    def check(self, expected, offsets, limit=0, max_size=0, fudge=0):
        coalesce = transport.Transport._coalesce_offsets
        exp = [transport._CoalescedOffset(*x) for x in expected]
        out = list(coalesce(offsets, limit=limit, fudge_factor=fudge,
                            max_size=max_size))
        self.assertEqual(exp, out)

    def test_coalesce_empty(self):
        self.check([], [])

    def test_coalesce_simple(self):
        self.check([(0, 10, [(0, 10)])], [(0, 10)])

    def test_coalesce_unrelated(self):
        self.check([(0, 10, [(0, 10)]),
                    (20, 10, [(0, 10)]),
                    ], [(0, 10), (20, 10)])

    def test_coalesce_unsorted(self):
        self.check([(20, 10, [(0, 10)]),
                    (0, 10, [(0, 10)]),
                    ], [(20, 10), (0, 10)])

    def test_coalesce_nearby(self):
        self.check([(0, 20, [(0, 10), (10, 10)])],
                   [(0, 10), (10, 10)])

    def test_coalesce_overlapped(self):
        self.assertRaises(ValueError,
                          self.check, [(0, 15, [(0, 10), (5, 10)])],
                          [(0, 10), (5, 10)])

    def test_coalesce_limit(self):
        self.check([(10, 50, [(0, 10), (10, 10), (20, 10),
                              (30, 10), (40, 10)]),
                    (60, 50, [(0, 10), (10, 10), (20, 10),
                              (30, 10), (40, 10)]),
                    ], [(10, 10), (20, 10), (30, 10), (40, 10),
                        (50, 10), (60, 10), (70, 10), (80, 10),
                        (90, 10), (100, 10)],
                   limit=5)

    def test_coalesce_no_limit(self):
        self.check([(10, 100, [(0, 10), (10, 10), (20, 10),
                               (30, 10), (40, 10), (50, 10),
                               (60, 10), (70, 10), (80, 10),
                               (90, 10)]),
                    ], [(10, 10), (20, 10), (30, 10), (40, 10),
                        (50, 10), (60, 10), (70, 10), (80, 10),
                        (90, 10), (100, 10)])

    def test_coalesce_fudge(self):
        self.check([(10, 30, [(0, 10), (20, 10)]),
                    (100, 10, [(0, 10)]),
                    ], [(10, 10), (30, 10), (100, 10)],
                   fudge=10)

    def test_coalesce_max_size(self):
        self.check([(10, 20, [(0, 10), (10, 10)]),
                    (30, 50, [(0, 50)]),
                    # If one range is above max_size, it gets its own coalesced
                    # offset
                    (100, 80, [(0, 80)]), ],
                   [(10, 10), (20, 10), (30, 50), (100, 80)],
                   max_size=50)

    def test_coalesce_no_max_size(self):
        self.check([(10, 170, [(0, 10), (10, 10), (20, 50), (70, 100)])],
                   [(10, 10), (20, 10), (30, 50), (80, 100)],
                   )

    def test_coalesce_default_limit(self):
        # By default we use a 100MB max size.
        ten_mb = 10 * 1024 * 1024
        self.check(
            [(0, 10 * ten_mb, [(i * ten_mb, ten_mb) for i in range(10)]),
             (10 * ten_mb, ten_mb, [(0, ten_mb)])],
            [(i * ten_mb, ten_mb) for i in range(11)])
        self.check(
            [(0, 11 * ten_mb, [(i * ten_mb, ten_mb) for i in range(11)])],
            [(i * ten_mb, ten_mb) for i in range(11)],
            max_size=1 * 1024 * 1024 * 1024)


class TestMemoryServer(tests.TestCase):

    def test_create_server(self):
        server = memory.MemoryServer()
        server.start_server()
        url = server.get_url()
        self.assertTrue(url in transport.transport_list_registry)
        t = transport.get_transport_from_url(url)
        del t
        server.stop_server()
        self.assertFalse(url in transport.transport_list_registry)
        self.assertRaises(UnsupportedProtocol,
                          transport.get_transport, url)


class TestMemoryTransport(tests.TestCase):

    def test_get_transport(self):
        memory.MemoryTransport()

    def test_clone(self):
        t = memory.MemoryTransport()
        self.assertTrue(isinstance(t, memory.MemoryTransport))
        self.assertEqual("memory:///", t.clone("/").base)

    def test_abspath(self):
        t = memory.MemoryTransport()
        self.assertEqual("memory:///relpath", t.abspath('relpath'))

    def test_abspath_of_root(self):
        t = memory.MemoryTransport()
        self.assertEqual("memory:///", t.base)
        self.assertEqual("memory:///", t.abspath('/'))

    def test_abspath_of_relpath_starting_at_root(self):
        t = memory.MemoryTransport()
        self.assertEqual("memory:///foo", t.abspath('/foo'))

    def test_append_and_get(self):
        t = memory.MemoryTransport()
        t.append_bytes('path', b'content')
        self.assertEqual(t.get('path').read(), b'content')
        t.append_file('path', BytesIO(b'content'))
        with t.get('path') as f:
            self.assertEqual(f.read(), b'contentcontent')

    def test_put_and_get(self):
        t = memory.MemoryTransport()
        t.put_file('path', BytesIO(b'content'))
        self.assertEqual(t.get('path').read(), b'content')
        t.put_bytes('path', b'content')
        self.assertEqual(t.get('path').read(), b'content')

    def test_append_without_dir_fails(self):
        t = memory.MemoryTransport()
        self.assertRaises(NoSuchFile,
                          t.append_bytes, 'dir/path', b'content')

    def test_put_without_dir_fails(self):
        t = memory.MemoryTransport()
        self.assertRaises(NoSuchFile,
                          t.put_file, 'dir/path', BytesIO(b'content'))

    def test_get_missing(self):
        transport = memory.MemoryTransport()
        self.assertRaises(NoSuchFile, transport.get, 'foo')

    def test_has_missing(self):
        t = memory.MemoryTransport()
        self.assertEqual(False, t.has('foo'))

    def test_has_present(self):
        t = memory.MemoryTransport()
        t.append_bytes('foo', b'content')
        self.assertEqual(True, t.has('foo'))

    def test_list_dir(self):
        t = memory.MemoryTransport()
        t.put_bytes('foo', b'content')
        t.mkdir('dir')
        t.put_bytes('dir/subfoo', b'content')
        t.put_bytes('dirlike', b'content')

        self.assertEqual(['dir', 'dirlike', 'foo'], sorted(t.list_dir('.')))
        self.assertEqual(['subfoo'], sorted(t.list_dir('dir')))

    def test_mkdir(self):
        t = memory.MemoryTransport()
        t.mkdir('dir')
        t.append_bytes('dir/path', b'content')
        with t.get('dir/path') as f:
            self.assertEqual(f.read(), b'content')

    def test_mkdir_missing_parent(self):
        t = memory.MemoryTransport()
        self.assertRaises(NoSuchFile, t.mkdir, 'dir/dir')

    def test_mkdir_twice(self):
        t = memory.MemoryTransport()
        t.mkdir('dir')
        self.assertRaises(FileExists, t.mkdir, 'dir')

    def test_parameters(self):
        t = memory.MemoryTransport()
        self.assertEqual(True, t.listable())
        self.assertEqual(False, t.is_readonly())

    def test_iter_files_recursive(self):
        t = memory.MemoryTransport()
        t.mkdir('dir')
        t.put_bytes('dir/foo', b'content')
        t.put_bytes('dir/bar', b'content')
        t.put_bytes('bar', b'content')
        paths = set(t.iter_files_recursive())
        self.assertEqual({'dir/foo', 'dir/bar', 'bar'}, paths)

    def test_stat(self):
        t = memory.MemoryTransport()
        t.put_bytes('foo', b'content')
        t.put_bytes('bar', b'phowar')
        self.assertEqual(7, t.stat('foo').st_size)
        self.assertEqual(6, t.stat('bar').st_size)


class ChrootDecoratorTransportTest(tests.TestCase):
    """Chroot decoration specific tests."""

    def test_abspath(self):
        # The abspath is always relative to the chroot_url.
        server = chroot.ChrootServer(
            transport.get_transport_from_url('memory:///foo/bar/'))
        self.start_server(server)
        t = transport.get_transport_from_url(server.get_url())
        self.assertEqual(server.get_url(), t.abspath('/'))

        subdir_t = t.clone('subdir')
        self.assertEqual(server.get_url(), subdir_t.abspath('/'))

    def test_clone(self):
        server = chroot.ChrootServer(
            transport.get_transport_from_url('memory:///foo/bar/'))
        self.start_server(server)
        t = transport.get_transport_from_url(server.get_url())
        # relpath from root and root path are the same
        relpath_cloned = t.clone('foo')
        abspath_cloned = t.clone('/foo')
        self.assertEqual(server, relpath_cloned.server)
        self.assertEqual(server, abspath_cloned.server)

    def test_chroot_url_preserves_chroot(self):
        """Calling get_transport on a chroot transport's base should produce a
        transport with exactly the same behaviour as the original chroot
        transport.

        This is so that it is not possible to escape a chroot by doing::
            url = chroot_transport.base
            parent_url = urlutils.join(url, '..')
            new_t = transport.get_transport_from_url(parent_url)
        """
        server = chroot.ChrootServer(
            transport.get_transport_from_url('memory:///path/subpath'))
        self.start_server(server)
        t = transport.get_transport_from_url(server.get_url())
        new_t = transport.get_transport_from_url(t.base)
        self.assertEqual(t.server, new_t.server)
        self.assertEqual(t.base, new_t.base)

    def test_urljoin_preserves_chroot(self):
        """Using urlutils.join(url, '..') on a chroot URL should not produce a
        URL that escapes the intended chroot.

        This is so that it is not possible to escape a chroot by doing::
            url = chroot_transport.base
            parent_url = urlutils.join(url, '..')
            new_t = transport.get_transport_from_url(parent_url)
        """
        server = chroot.ChrootServer(
            transport.get_transport_from_url('memory:///path/'))
        self.start_server(server)
        t = transport.get_transport_from_url(server.get_url())
        self.assertRaises(
            urlutils.InvalidURLJoin, urlutils.join, t.base, '..')


class TestChrootServer(tests.TestCase):

    def test_construct(self):
        backing_transport = memory.MemoryTransport()
        server = chroot.ChrootServer(backing_transport)
        self.assertEqual(backing_transport, server.backing_transport)

    def test_setUp(self):
        backing_transport = memory.MemoryTransport()
        server = chroot.ChrootServer(backing_transport)
        server.start_server()
        self.addCleanup(server.stop_server)
        self.assertTrue(server.scheme
                        in transport._get_protocol_handlers().keys())

    def test_stop_server(self):
        backing_transport = memory.MemoryTransport()
        server = chroot.ChrootServer(backing_transport)
        server.start_server()
        server.stop_server()
        self.assertFalse(server.scheme
                         in transport._get_protocol_handlers().keys())

    def test_get_url(self):
        backing_transport = memory.MemoryTransport()
        server = chroot.ChrootServer(backing_transport)
        server.start_server()
        self.addCleanup(server.stop_server)
        self.assertEqual('chroot-%d:///' % id(server), server.get_url())


class TestHooks(tests.TestCase):
    """Basic tests for transport hooks"""

    def _get_connected_transport(self):
        return transport.ConnectedTransport("bogus:nowhere")

    def test_transporthooks_initialisation(self):
        """Check all expected transport hook points are set up"""
        hookpoint = transport.TransportHooks()
        self.assertTrue("post_connect" in hookpoint,
                        "post_connect not in {}".format(hookpoint))

    def test_post_connect(self):
        """Ensure the post_connect hook is called when _set_transport is"""
        calls = []
        transport.Transport.hooks.install_named_hook("post_connect",
                                                     calls.append, None)
        t = self._get_connected_transport()
        self.assertLength(0, calls)
        t._set_connection("connection", "auth")
        self.assertEqual(calls, [t])


class PathFilteringDecoratorTransportTest(tests.TestCase):
    """Pathfilter decoration specific tests."""

    def test_abspath(self):
        # The abspath is always relative to the base of the backing transport.
        server = pathfilter.PathFilteringServer(
            transport.get_transport_from_url('memory:///foo/bar/'),
            lambda x: x)
        server.start_server()
        t = transport.get_transport_from_url(server.get_url())
        self.assertEqual(server.get_url(), t.abspath('/'))

        subdir_t = t.clone('subdir')
        self.assertEqual(server.get_url(), subdir_t.abspath('/'))
        server.stop_server()

    def make_pf_transport(self, filter_func=None):
        """Make a PathFilteringTransport backed by a MemoryTransport.

        :param filter_func: by default this will be a no-op function.  Use this
            parameter to override it."""
        if filter_func is None:
            def filter_func(x):
                return x
        server = pathfilter.PathFilteringServer(
            transport.get_transport_from_url('memory:///foo/bar/'),
            filter_func)
        server.start_server()
        self.addCleanup(server.stop_server)
        return transport.get_transport_from_url(server.get_url())

    def test__filter(self):
        # _filter (with an identity func as filter_func) always returns
        # paths relative to the base of the backing transport.
        t = self.make_pf_transport()
        self.assertEqual('foo', t._filter('foo'))
        self.assertEqual('foo/bar', t._filter('foo/bar'))
        self.assertEqual('', t._filter('..'))
        self.assertEqual('', t._filter('/'))
        # The base of the pathfiltering transport is taken into account too.
        t = t.clone('subdir1/subdir2')
        self.assertEqual('subdir1/subdir2/foo', t._filter('foo'))
        self.assertEqual('subdir1/subdir2/foo/bar', t._filter('foo/bar'))
        self.assertEqual('subdir1', t._filter('..'))
        self.assertEqual('', t._filter('/'))

    def test_filter_invocation(self):
        filter_log = []

        def filter(path):
            filter_log.append(path)
            return path
        t = self.make_pf_transport(filter)
        t.has('abc')
        self.assertEqual(['abc'], filter_log)
        del filter_log[:]
        t.clone('abc').has('xyz')
        self.assertEqual(['abc/xyz'], filter_log)
        del filter_log[:]
        t.has('/abc')
        self.assertEqual(['abc'], filter_log)

    def test_clone(self):
        t = self.make_pf_transport()
        # relpath from root and root path are the same
        relpath_cloned = t.clone('foo')
        abspath_cloned = t.clone('/foo')
        self.assertEqual(t.server, relpath_cloned.server)
        self.assertEqual(t.server, abspath_cloned.server)

    def test_url_preserves_pathfiltering(self):
        """Calling get_transport on a pathfiltered transport's base should
        produce a transport with exactly the same behaviour as the original
        pathfiltered transport.

        This is so that it is not possible to escape (accidentally or
        otherwise) the filtering by doing::
            url = filtered_transport.base
            parent_url = urlutils.join(url, '..')
            new_t = transport.get_transport_from_url(parent_url)
        """
        t = self.make_pf_transport()
        new_t = transport.get_transport_from_url(t.base)
        self.assertEqual(t.server, new_t.server)
        self.assertEqual(t.base, new_t.base)


class ReadonlyDecoratorTransportTest(tests.TestCase):
    """Readonly decoration specific tests."""

    def test_local_parameters(self):
        # connect to . in readonly mode
        t = readonly.ReadonlyTransportDecorator('readonly+.')
        self.assertEqual(True, t.listable())
        self.assertEqual(True, t.is_readonly())

    def test_http_parameters(self):
        from breezy.tests.http_server import HttpServer

        # connect to '.' via http which is not listable
        server = HttpServer()
        self.start_server(server)
        t = transport.get_transport_from_url('readonly+' + server.get_url())
        self.assertIsInstance(t, readonly.ReadonlyTransportDecorator)
        self.assertEqual(False, t.listable())
        self.assertEqual(True, t.is_readonly())


class FakeNFSDecoratorTests(tests.TestCaseInTempDir):
    """NFS decorator specific tests."""

    def get_nfs_transport(self, url):
        # connect to url with nfs decoration
        return fakenfs.FakeNFSTransportDecorator('fakenfs+' + url)

    def test_local_parameters(self):
        # the listable and is_readonly parameters
        # are not changed by the fakenfs decorator
        t = self.get_nfs_transport('.')
        self.assertEqual(True, t.listable())
        self.assertEqual(False, t.is_readonly())

    def test_http_parameters(self):
        # the listable and is_readonly parameters
        # are not changed by the fakenfs decorator
        from breezy.tests.http_server import HttpServer

        # connect to '.' via http which is not listable
        server = HttpServer()
        self.start_server(server)
        t = self.get_nfs_transport(server.get_url())
        self.assertIsInstance(t, fakenfs.FakeNFSTransportDecorator)
        self.assertEqual(False, t.listable())
        self.assertEqual(True, t.is_readonly())

    def test_fakenfs_server_default(self):
        # a FakeNFSServer() should bring up a local relpath server for itself
        server = test_server.FakeNFSServer()
        self.start_server(server)
        # the url should be decorated appropriately
        self.assertStartsWith(server.get_url(), 'fakenfs+')
        # and we should be able to get a transport for it
        t = transport.get_transport_from_url(server.get_url())
        # which must be a FakeNFSTransportDecorator instance.
        self.assertIsInstance(t, fakenfs.FakeNFSTransportDecorator)

    def test_fakenfs_rename_semantics(self):
        # a FakeNFS transport must mangle the way rename errors occur to
        # look like NFS problems.
        t = self.get_nfs_transport('.')
        self.build_tree(['from/', 'from/foo', 'to/', 'to/bar'],
                        transport=t)
        self.assertRaises(errors.ResourceBusy, t.rename, 'from', 'to')


class FakeVFATDecoratorTests(tests.TestCaseInTempDir):
    """Tests for simulation of VFAT restrictions"""

    def get_vfat_transport(self, url):
        """Return vfat-backed transport for test directory"""
        from breezy.transport.fakevfat import FakeVFATTransportDecorator
        return FakeVFATTransportDecorator('vfat+' + url)

    def test_transport_creation(self):
        from breezy.transport.fakevfat import FakeVFATTransportDecorator
        t = self.get_vfat_transport('.')
        self.assertIsInstance(t, FakeVFATTransportDecorator)

    def test_transport_mkdir(self):
        t = self.get_vfat_transport('.')
        t.mkdir('HELLO')
        self.assertTrue(t.has('hello'))
        self.assertTrue(t.has('Hello'))

    def test_forbidden_chars(self):
        t = self.get_vfat_transport('.')
        self.assertRaises(ValueError, t.has, "<NU>")


class BadTransportHandler(transport.Transport):
    def __init__(self, base_url):
        raise errors.DependencyNotPresent('some_lib',
                                          'testing missing dependency')


class BackupTransportHandler(transport.Transport):
    """Test transport that works as a backup for the BadTransportHandler"""
    pass


class TestTransportImplementation(tests.TestCaseInTempDir):
    """Implementation verification for transports.

    To verify a transport we need a server factory, which is a callable
    that accepts no parameters and returns an implementation of
    breezy.transport.Server.

    That Server is then used to construct transport instances and test
    the transport via loopback activity.

    Currently this assumes that the Transport object is connected to the
    current working directory.  So that whatever is done
    through the transport, should show up in the working
    directory, and vice-versa. This is a bug, because its possible to have
    URL schemes which provide access to something that may not be
    result in storage on the local disk, i.e. due to file system limits, or
    due to it being a database or some other non-filesystem tool.

    This also tests to make sure that the functions work with both
    generators and lists (assuming iter(list) is effectively a generator)
    """

    def setUp(self):
        super().setUp()
        self._server = self.transport_server()
        self.start_server(self._server)

    def get_transport(self, relpath=None):
        """Return a connected transport to the local directory.

        :param relpath: a path relative to the base url.
        """
        base_url = self._server.get_url()
        url = self._adjust_url(base_url, relpath)
        # try getting the transport via the regular interface:
        t = transport.get_transport_from_url(url)
        # vila--20070607 if the following are commented out the test suite
        # still pass. Is this really still needed or was it a forgotten
        # temporary fix ?
        if not isinstance(t, self.transport_class):
            # we did not get the correct transport class type. Override the
            # regular connection behaviour by direct construction.
            t = self.transport_class(url)
        return t


class TestTransportFromPath(tests.TestCaseInTempDir):

    def test_with_path(self):
        t = transport.get_transport_from_path(self.test_dir)
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(t.base.rstrip("/"),
                         urlutils.local_path_to_url(self.test_dir))

    def test_with_url(self):
        t = transport.get_transport_from_path("file:")
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(
            t.base.rstrip("/"),
            urlutils.local_path_to_url(os.path.join(self.test_dir, "file:")))


class TestTransportFromUrl(tests.TestCaseInTempDir):

    def test_with_path(self):
        self.assertRaises(urlutils.InvalidURL, transport.get_transport_from_url,
                          self.test_dir)

    def test_with_url(self):
        url = urlutils.local_path_to_url(self.test_dir)
        t = transport.get_transport_from_url(url)
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(t.base.rstrip("/"), url)

    def test_with_url_and_segment_parameters(self):
        url = urlutils.local_path_to_url(self.test_dir) + ",branch=foo"
        t = transport.get_transport_from_url(url)
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(t.base.rstrip("/"), url)
        with open(os.path.join(self.test_dir, "afile"), 'w') as f:
            f.write("data")
        self.assertTrue(t.has("afile"))


class TestLocalTransports(tests.TestCase):

    def test_get_transport_from_abspath(self):
        here = osutils.abspath('.')
        t = transport.get_transport(here)
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(t.base, urlutils.local_path_to_url(here) + '/')

    def test_get_transport_from_relpath(self):
        t = transport.get_transport('.')
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(t.base, urlutils.local_path_to_url('.') + '/')

    def test_get_transport_from_local_url(self):
        here = osutils.abspath('.')
        here_url = urlutils.local_path_to_url(here) + '/'
        t = transport.get_transport(here_url)
        self.assertIsInstance(t, local.LocalTransport)
        self.assertEqual(t.base, here_url)

    def test_local_abspath(self):
        here = osutils.abspath('.')
        t = transport.get_transport(here)
        self.assertEqual(t.local_abspath(''), here)


class TestLocalTransportMutation(tests.TestCaseInTempDir):

    def test_local_transport_mkdir(self):
        here = osutils.abspath('.')
        t = transport.get_transport(here)
        t.mkdir('test')
        self.assertTrue(os.path.exists('test'))

    def test_local_transport_mkdir_permission_denied(self):
        # See https://bugs.launchpad.net/bzr/+bug/606537
        here = osutils.abspath('.')
        t = transport.get_transport(here)

        def fake_chmod(path, mode):
            e = OSError('permission denied')
            e.errno = errno.EPERM
            raise e
        self.overrideAttr(os, 'chmod', fake_chmod)
        t.mkdir('test')
        t.mkdir('test2', mode=0o707)
        self.assertTrue(os.path.exists('test'))
        self.assertTrue(os.path.exists('test2'))


class TestLocalTransportWriteStream(tests.TestCaseWithTransport):

    def test_local_fdatasync_calls_fdatasync(self):
        """Check fdatasync on a stream tries to flush the data to the OS.

        We can't easily observe the external effect but we can at least see
        it's called.
        """
        sentinel = object()
        fdatasync = getattr(os, 'fdatasync', sentinel)
        if fdatasync is sentinel:
            raise tests.TestNotApplicable('fdatasync not supported')
        t = self.get_transport('.')
        calls = self.recordCalls(os, 'fdatasync')
        w = t.open_write_stream('out')
        w.write(b'foo')
        w.fdatasync()
        with open('out', 'rb') as f:
            # Should have been flushed.
            self.assertEqual(f.read(), b'foo')
        self.assertEqual(len(calls), 1, calls)

    def test_missing_directory(self):
        t = self.get_transport('.')
        self.assertRaises(NoSuchFile, t.open_write_stream, 'dir/foo')


class TestWin32LocalTransport(tests.TestCase):

    def test_unc_clone_to_root(self):
        self.requireFeature(features.win32_feature)
        # Win32 UNC path like \\HOST\path
        # clone to root should stop at least at \\HOST part
        # not on \\
        t = local.EmulatedWin32LocalTransport('file://HOST/path/to/some/dir/')
        for i in range(4):
            t = t.clone('..')
        self.assertEqual(t.base, 'file://HOST/')
        # make sure we reach the root
        t = t.clone('..')
        self.assertEqual(t.base, 'file://HOST/')


class TestConnectedTransport(tests.TestCase):
    """Tests for connected to remote server transports"""

    def test_parse_url(self):
        t = transport.ConnectedTransport(
            'http://simple.example.com/home/source')
        self.assertEqual(t._parsed_url.host, 'simple.example.com')
        self.assertEqual(t._parsed_url.port, None)
        self.assertEqual(t._parsed_url.path, '/home/source/')
        self.assertTrue(t._parsed_url.user is None)
        self.assertTrue(t._parsed_url.password is None)

        self.assertEqual(t.base, 'http://simple.example.com/home/source/')

    def test_parse_url_with_at_in_user(self):
        # Bug 228058
        t = transport.ConnectedTransport('ftp://user@host.com@www.host.com/')
        self.assertEqual(t._parsed_url.user, 'user@host.com')

    def test_parse_quoted_url(self):
        t = transport.ConnectedTransport(
            'http://ro%62ey:h%40t@ex%41mple.com:2222/path')
        self.assertEqual(t._parsed_url.host, 'exAmple.com')
        self.assertEqual(t._parsed_url.port, 2222)
        self.assertEqual(t._parsed_url.user, 'robey')
        self.assertEqual(t._parsed_url.password, 'h@t')
        self.assertEqual(t._parsed_url.path, '/path/')

        # Base should not keep track of the password
        self.assertEqual(t.base, 'http://ro%62ey@ex%41mple.com:2222/path/')

    def test_parse_invalid_url(self):
        self.assertRaises(urlutils.InvalidURL,
                          transport.ConnectedTransport,
                          'sftp://lily.org:~janneke/public/bzr/gub')

    def test_relpath(self):
        t = transport.ConnectedTransport('sftp://user@host.com/abs/path')

        self.assertEqual(t.relpath('sftp://user@host.com/abs/path/sub'),
                         'sub')
        self.assertRaises(errors.PathNotChild, t.relpath,
                          'http://user@host.com/abs/path/sub')
        self.assertRaises(errors.PathNotChild, t.relpath,
                          'sftp://user2@host.com/abs/path/sub')
        self.assertRaises(errors.PathNotChild, t.relpath,
                          'sftp://user@otherhost.com/abs/path/sub')
        self.assertRaises(errors.PathNotChild, t.relpath,
                          'sftp://user@host.com:33/abs/path/sub')
        # Make sure it works when we don't supply a username
        t = transport.ConnectedTransport('sftp://host.com/abs/path')
        self.assertEqual(t.relpath('sftp://host.com/abs/path/sub'), 'sub')

        # Make sure it works when parts of the path will be url encoded
        t = transport.ConnectedTransport('sftp://host.com/dev/%path')
        self.assertEqual(t.relpath('sftp://host.com/dev/%path/sub'), 'sub')

    def test_connection_sharing_propagate_credentials(self):
        t = transport.ConnectedTransport('ftp://user@host.com/abs/path')
        self.assertEqual('user', t._parsed_url.user)
        self.assertEqual('host.com', t._parsed_url.host)
        self.assertIs(None, t._get_connection())
        self.assertIs(None, t._parsed_url.password)
        c = t.clone('subdir')
        self.assertIs(None, c._get_connection())
        self.assertIs(None, t._parsed_url.password)

        # Simulate the user entering a password
        password = 'secret'
        connection = object()
        t._set_connection(connection, password)
        self.assertIs(connection, t._get_connection())
        self.assertIs(password, t._get_credentials())
        self.assertIs(connection, c._get_connection())
        self.assertIs(password, c._get_credentials())

        # credentials can be updated
        new_password = 'even more secret'
        c._update_credentials(new_password)
        self.assertIs(connection, t._get_connection())
        self.assertIs(new_password, t._get_credentials())
        self.assertIs(connection, c._get_connection())
        self.assertIs(new_password, c._get_credentials())


class TestReusedTransports(tests.TestCase):
    """Tests for transport reuse"""

    def test_reuse_same_transport(self):
        possible_transports = []
        t1 = transport.get_transport_from_url(
            'http://foo/', possible_transports=possible_transports)
        self.assertEqual([t1], possible_transports)
        t2 = transport.get_transport_from_url('http://foo/',
                                              possible_transports=[t1])
        self.assertIs(t1, t2)

        # Also check that final '/' are handled correctly
        t3 = transport.get_transport_from_url('http://foo/path/')
        t4 = transport.get_transport_from_url('http://foo/path',
                                              possible_transports=[t3])
        self.assertIs(t3, t4)

        t5 = transport.get_transport_from_url('http://foo/path')
        t6 = transport.get_transport_from_url('http://foo/path/',
                                              possible_transports=[t5])
        self.assertIs(t5, t6)

    def test_don_t_reuse_different_transport(self):
        t1 = transport.get_transport_from_url('http://foo/path')
        t2 = transport.get_transport_from_url('http://bar/path',
                                              possible_transports=[t1])
        self.assertIsNot(t1, t2)


class TestTransportTrace(tests.TestCase):

    def test_decorator(self):
        t = transport.get_transport_from_url('trace+memory://')
        self.assertIsInstance(
            t, breezy.transport.trace.TransportTraceDecorator)

    def test_clone_preserves_activity(self):
        t = transport.get_transport_from_url('trace+memory://')
        t2 = t.clone('.')
        self.assertTrue(t is not t2)
        self.assertTrue(t._activity is t2._activity)

    # the following specific tests are for the operations that have made use of
    # logging in tests; we could test every single operation but doing that
    # still won't cause a test failure when the top level Transport API
    # changes; so there is little return doing that.
    def test_get(self):
        t = transport.get_transport_from_url('trace+memory:///')
        t.put_bytes('foo', b'barish')
        t.get('foo')
        expected_result = []
        # put_bytes records the bytes, not the content to avoid memory
        # pressure.
        expected_result.append(('put_bytes', 'foo', 6, None))
        # get records the file name only.
        expected_result.append(('get', 'foo'))
        self.assertEqual(expected_result, t._activity)

    def test_readv(self):
        t = transport.get_transport_from_url('trace+memory:///')
        t.put_bytes('foo', b'barish')
        list(t.readv('foo', [(0, 1), (3, 2)],
                     adjust_for_latency=True, upper_limit=6))
        expected_result = []
        # put_bytes records the bytes, not the content to avoid memory
        # pressure.
        expected_result.append(('put_bytes', 'foo', 6, None))
        # readv records the supplied offset request
        expected_result.append(('readv', 'foo', [(0, 1), (3, 2)], True, 6))
        self.assertEqual(expected_result, t._activity)


class TestSSHConnections(tests.TestCaseWithTransport):

    def test_bzr_connect_to_bzr_ssh(self):
        """get_transport of a bzr+ssh:// behaves correctly.

        bzr+ssh:// should cause bzr to run a remote bzr smart server over SSH.
        """
        # This test actually causes a bzr instance to be invoked, which is very
        # expensive: it should be the only such test in the test suite.
        # A reasonable evolution for this would be to simply check inside
        # check_channel_exec_request that the command is appropriate, and then
        # satisfy requests in-process.
        self.requireFeature(features.paramiko)
        # SFTPFullAbsoluteServer has a get_url method, and doesn't
        # override the interface (doesn't change self._vendor).
        # Note that this does encryption, so can be slow.
        from breezy.tests import stub_sftp

        # Start an SSH server
        self.command_executed = []
        # XXX: This is horrible -- we define a really dumb SSH server that
        # executes commands, and manage the hooking up of stdin/out/err to the
        # SSH channel ourselves.  Surely this has already been implemented
        # elsewhere?
        started = []

        class StubSSHServer(stub_sftp.StubServer):

            test = self

            def check_channel_exec_request(self, channel, command):
                self.test.command_executed.append(command)
                proc = subprocess.Popen(
                    command, shell=True, stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                    bufsize=0)

                # XXX: horribly inefficient, not to mention ugly.
                # Start a thread for each of stdin/out/err, and relay bytes
                # from the subprocess to channel and vice versa.
                def ferry_bytes(read, write, close):
                    while True:
                        bytes = read(1)
                        if bytes == b'':
                            close()
                            break
                        write(bytes)

                file_functions = [
                    (channel.recv, proc.stdin.write, proc.stdin.close),
                    (proc.stdout.read, channel.sendall, channel.close),
                    (proc.stderr.read, channel.sendall_stderr, channel.close)]
                started.append(proc)
                for read, write, close in file_functions:
                    t = threading.Thread(
                        target=ferry_bytes, args=(read, write, close))
                    t.start()
                    started.append(t)

                return True

        ssh_server = stub_sftp.SFTPFullAbsoluteServer(StubSSHServer)
        # We *don't* want to override the default SSH vendor: the detected one
        # is the one to use.

        # FIXME: I don't understand the above comment, SFTPFullAbsoluteServer
        # inherits from SFTPServer which forces the SSH vendor to
        # ssh.ParamikoVendor(). So it's forced, not detected. --vila 20100623
        self.start_server(ssh_server)
        port = ssh_server.port

        bzr_remote_command = self.get_brz_command()
        self.overrideEnv('BZR_REMOTE_PATH', ' '.join(bzr_remote_command))
        self.overrideEnv('PYTHONPATH', ':'.join(sys.path))

        # Access the branch via a bzr+ssh URL.  The BZR_REMOTE_PATH environment
        # variable is used to tell bzr what command to run on the remote end.
        path_to_branch = osutils.abspath('.')
        if sys.platform == 'win32':
            # On Windows, we export all drives as '/C:/, etc. So we need to
            # prefix a '/' to get the right path.
            path_to_branch = '/' + path_to_branch
        url = 'bzr+ssh://fred:secret@localhost:%d%s' % (port, path_to_branch)
        t = transport.get_transport(url)
        self.permit_url(t.base)
        t.mkdir('foo')

        self.assertEqual(
            [b'%s serve --inet --directory=/ --allow-writes' %
                ' '.join(bzr_remote_command).encode()],
            self.command_executed)
        # Make sure to disconnect, so that the remote process can stop, and we
        # can cleanup. Then pause the test until everything is shutdown
        t._client._medium.disconnect()
        if not started:
            return
        # First wait for the subprocess
        started[0].wait()
        # And the rest are threads
        for t in started[1:]:
            t.join()



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black box tests for the upgrade ui."""
import os
import stat

from breezy import bzr, controldir, lockdir, ui, urlutils
from breezy.bzr import bzrdir
from breezy.bzr.knitpack_repo import RepositoryFormatKnitPack1
from breezy.tests import TestCaseWithTransport, features
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer


class OldBzrDir(bzrdir.BzrDirMeta1):
    """An test brz dir implementation"""

    def needs_format_conversion(self, format):
        return not isinstance(format, self.__class__)


class ConvertOldTestToMeta(controldir.Converter):
    """A trivial converter, used for testing."""

    def convert(self, to_convert, pb):
        ui.ui_factory.note('starting upgrade from old test format to 2a')
        to_convert.control_transport.put_bytes(
            'branch-format',
            bzrdir.BzrDirMetaFormat1().get_format_string(),
            mode=to_convert._get_file_mode())
        return controldir.ControlDir.open(to_convert.user_url)


class OldBzrDirFormat(bzrdir.BzrDirMetaFormat1):

    _lock_class = lockdir.LockDir

    def get_converter(self, format=None):
        return ConvertOldTestToMeta()

    @classmethod
    def get_format_string(cls):
        return b"Ancient Test Format"

    def _open(self, transport):
        return OldBzrDir(transport, self)


class TestWithUpgradableBranches(TestCaseWithTransport):

    def setUp(self):
        super().setUp()

    def make_current_format_branch_and_checkout(self):
        current_tree = self.make_branch_and_tree('current_format_branch',
                                                 format='default')
        current_tree.branch.create_checkout(
            self.get_url('current_format_checkout'), lightweight=True)

    def test_readonly_url_error(self):
        self.make_branch_and_tree("old_format_branch", format="knit")
        (out, err) = self.run_bzr(
            ['upgrade', self.get_readonly_url("old_format_branch")], retcode=3)
        err_msg = 'Upgrade URL cannot work with readonly URLs.'
        self.assertEqualDiff('conversion error: %s\nbrz: ERROR: %s\n'
                             % (err_msg, err_msg),
                             err)

    def test_upgrade_up_to_date(self):
        self.make_current_format_branch_and_checkout()
        # when up to date we should get a message to that effect
        burl = self.get_transport('current_format_branch').local_abspath(".")
        (out, err) = self.run_bzr('upgrade current_format_branch', retcode=0)
        self.assertEqual(
            'Upgrading branch %s/ ...\n'
            'The branch format %s is already at the most recent format.\n'
            % (burl, 'Meta directory format 1'),
            out)

    def test_upgrade_up_to_date_checkout_warns_branch_left_alone(self):
        self.make_current_format_branch_and_checkout()
        # when upgrading a checkout, the branch location and a suggestion
        # to upgrade it should be emitted even if the checkout is up to
        # date
        burl = self.get_transport('current_format_branch').local_abspath(".")
        curl = self.get_transport('current_format_checkout').local_abspath(".")
        (out, err) = self.run_bzr('upgrade current_format_checkout', retcode=0)
        self.assertEqual(
            'Upgrading branch %s/ ...\nThis is a checkout.'
            ' The branch (%s/) needs to be upgraded separately.\n'
            'The branch format %s is already at the most recent format.\n'
            % (curl, burl, 'Meta directory format 1'),
            out)

    def test_upgrade_checkout(self):
        # upgrading a checkout should work
        pass

    def test_upgrade_repository_scans_branches(self):
        # we should get individual upgrade notes for each branch even the
        # anonymous branch
        pass

    def test_upgrade_branch_in_repo(self):
        # upgrading a branch in a repo should warn about not upgrading the repo
        pass

    def test_upgrade_control_dir(self):
        old_format = OldBzrDirFormat()
        self.addCleanup(bzr.BzrProber.formats.remove,
                        old_format.get_format_string())
        bzr.BzrProber.formats.register(old_format.get_format_string(),
                                       old_format)
        self.addCleanup(controldir.ControlDirFormat._set_default_format,
                        controldir.ControlDirFormat.get_default_format())

        # setup an old format branch we can upgrade from.
        path = 'old_format_branch'
        self.make_branch_and_tree(path, format=old_format)
        transport = self.get_transport(path)
        url = transport.base
        display_url = transport.local_abspath('.')
        # check --format takes effect
        controldir.ControlDirFormat._set_default_format(old_format)
        backup_dir = 'backup.bzr.~1~'
        (out, err) = self.run_bzr(
            ['upgrade', '--format=2a', url])
        self.assertEqualDiff("""Upgrading branch {}/ ...
starting upgrade of {}/
making backup of {}/.bzr
  to {}/{}
starting upgrade from old test format to 2a
finished
""".format(display_url, display_url, display_url, display_url, backup_dir), out)
        self.assertEqualDiff("", err)
        self.assertTrue(isinstance(
            controldir.ControlDir.open(self.get_url(path))._format,
            bzrdir.BzrDirMetaFormat1))

    def test_upgrade_explicit_knit(self):
        # users can force an upgrade to knit format from a metadir pack 0.92
        # branch to a 2a branch.
        self.make_branch_and_tree('branch', format='knit')
        transport = self.get_transport('branch')
        url = transport.base
        display_url = transport.local_abspath('.')
        # check --format takes effect
        backup_dir = 'backup.bzr.~1~'
        (out, err) = self.run_bzr(
            ['upgrade', '--format=pack-0.92', url])
        self.assertEqualDiff("""Upgrading branch {}/ ...
starting upgrade of {}/
making backup of {}/.bzr
  to {}/{}
starting repository conversion
repository converted
finished
""".format(display_url, display_url, display_url, display_url, backup_dir),
            out)
        self.assertEqualDiff("", err)
        converted_dir = controldir.ControlDir.open(self.get_url('branch'))
        self.assertTrue(isinstance(converted_dir._format,
                                   bzrdir.BzrDirMetaFormat1))
        self.assertTrue(isinstance(converted_dir.open_repository()._format,
                                   RepositoryFormatKnitPack1))

    def test_upgrade_repo(self):
        self.run_bzr('init-shared-repository --format=pack-0.92 repo')
        self.run_bzr('upgrade --format=2a repo')

    def assertLegalOption(self, option_str):
        # Confirm that an option is legal. (Lower level tests are
        # expected to validate the actual functionality.)
        self.run_bzr('init --format=pack-0.92 branch-foo')
        self.run_bzr('upgrade --format=2a branch-foo {}'.format(option_str))

    def assertBranchFormat(self, dir, format):
        branch = controldir.ControlDir.open_tree_or_branch(self.get_url(dir))[
            1]
        branch_format = branch._format
        meta_format = controldir.format_registry.make_controldir(format)
        expected_format = meta_format.get_branch_format()
        self.assertEqual(expected_format, branch_format)

    def test_upgrade_clean_supported(self):
        self.assertLegalOption('--clean')
        self.assertBranchFormat('branch-foo', '2a')
        backup_bzr_dir = os.path.join("branch-foo", "backup.bzr.~1~")
        self.assertFalse(os.path.exists(backup_bzr_dir))

    def test_upgrade_dry_run_supported(self):
        self.assertLegalOption('--dry-run')
        self.assertBranchFormat('branch-foo', 'pack-0.92')

    def test_upgrade_permission_check(self):
        """'backup.bzr' should retain permissions of .bzr. Bug #262450"""
        self.requireFeature(features.posix_permissions_feature)
        old_perms = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        backup_dir = 'backup.bzr.~1~'
        self.run_bzr('init --format=1.6')
        os.chmod('.bzr', old_perms)
        self.run_bzr('upgrade')
        new_perms = os.stat(backup_dir).st_mode & 0o777
        self.assertTrue(new_perms == old_perms)

    def test_upgrade_with_existing_backup_dir(self):
        self.make_branch_and_tree("old_format_branch", format="knit")
        t = self.get_transport("old_format_branch")
        url = t.base
        display_url = t.local_abspath('.')
        backup_dir1 = 'backup.bzr.~1~'
        backup_dir2 = 'backup.bzr.~2~'
        # explicitly create backup_dir1. brz should create the .~2~ directory
        # as backup
        t.mkdir(backup_dir1)
        (out, err) = self.run_bzr(
            ['upgrade', '--format=2a', url])
        self.assertEqualDiff("""Upgrading branch {}/ ...
starting upgrade of {}/
making backup of {}/.bzr
  to {}/{}
starting repository conversion
repository converted
finished
""".format(display_url, display_url, display_url, display_url, backup_dir2), out)
        self.assertEqualDiff("", err)
        self.assertTrue(isinstance(
            controldir.ControlDir.open(
                self.get_url("old_format_branch"))._format,
            bzrdir.BzrDirMetaFormat1))
        self.assertTrue(t.has(backup_dir2))


class SFTPTests(TestCaseWithSFTPServer):
    """Tests for upgrade over sftp."""

    def test_upgrade_url(self):
        self.run_bzr('init --format=pack-0.92')
        t = self.get_transport()
        url = t.base
        display_url = urlutils.unescape_for_display(url,
                                                    'utf-8')
        out, err = self.run_bzr(['upgrade', '--format=2a', url])
        backup_dir = 'backup.bzr.~1~'
        self.assertEqualDiff("""Upgrading branch {} ...
starting upgrade of {}
making backup of {}.bzr
  to {}{}
starting repository conversion
repository converted
finished
""".format(display_url, display_url, display_url, display_url, backup_dir), out)
        self.assertEqual('', err)


class UpgradeRecommendedTests(TestCaseWithTransport):

    def test_recommend_upgrade_wt4(self):
        # using a deprecated format gives a warning
        self.run_bzr('init --format=knit a')
        out, err = self.run_bzr('status a')
        self.assertContainsRe(err, 'brz upgrade .*[/\\\\]a')

    def test_no_upgrade_recommendation_from_bzrdir(self):
        # we should only get a recommendation to upgrade when we're accessing
        # the actual workingtree, not when we only open a bzrdir that contains
        # an old workngtree
        self.run_bzr('init --format=knit a')
        out, err = self.run_bzr('revno a')
        if err.find('upgrade') > -1:
            self.fail("message shouldn't suggest upgrade:\n%s" % err)

    def test_upgrade_shared_repo(self):
        repo = self.make_repository('repo', format='2a', shared=True)
        branch = self.make_branch_and_tree('repo/branch', format="pack-0.92")
        self.get_transport('repo/branch/.bzr/repository').delete_tree('.')
        out, err = self.run_bzr(['upgrade'], working_dir='repo/branch')



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.tests.per_repository import TestCaseWithRepository


class TestDefaultStackingPolicy(TestCaseWithRepository):

    def test_sprout_to_smart_server_stacking_policy_handling(self):
        """Obey policy where possible, ignore otherwise."""
        stack_on = self.make_branch('stack-on')
        parent_bzrdir = self.make_controldir('.', format='default')
        parent_bzrdir.get_config().set_default_stack_on('stack-on')
        source = self.make_branch('source')
        url = self.make_smart_server('target').abspath('')
        target = source.controldir.sprout(url).open_branch()
        self.assertEqual('../stack-on', target.get_stacked_on_url())
        self.assertEqual(
            source._format.network_name(), target._format.network_name())



# Copyright (C) 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black box tests for the reconcile command."""


from breezy import controldir, tests
from breezy.bzr import inventory
from breezy.repository import WriteGroup


class TrivialTest(tests.TestCaseWithTransport):

    def test_trivial_reconcile(self):
        t = controldir.ControlDir.create_standalone_workingtree('.')
        (out, err) = self.run_bzr('reconcile')
        if t.branch.repository._reconcile_backsup_inventory:
            does_backup_text = "Inventory ok.\n"
        else:
            does_backup_text = ""
        self.assertEqualDiff(out, "Reconciling branch %s\n"
                                  "revision_history ok.\n"
                                  "Reconciling repository %s\n"
                                  "%s"
                                  "Reconciliation complete.\n" %
                                  (t.branch.base,
                                   t.controldir.root_transport.base,
                                   does_backup_text))
        self.assertEqualDiff(err, "")

    def test_does_something_reconcile(self):
        t = controldir.ControlDir.create_standalone_workingtree('.')
        # an empty inventory with no revision will trigger reconciliation.
        repo = t.branch.repository
        inv = inventory.Inventory(revision_id=b'missing')
        inv.root.revision = b'missing'
        repo.lock_write()
        with repo.lock_write(), WriteGroup(repo):
            repo.add_inventory(b'missing', inv, [])
        (out, err) = self.run_bzr('reconcile')
        if repo._reconcile_backsup_inventory:
            does_backup_text = (
                "Backup Inventory created.\n"
                "Inventory regenerated.\n")
        else:
            does_backup_text = ""
        expected = ("Reconciling branch %s\n"
                    "revision_history ok.\n"
                    "Reconciling repository %s\n"
                    "%s"
                    "Reconciliation complete.\n" %
                    (t.branch.base,
                     t.controldir.root_transport.base,
                     does_backup_text))
        self.assertEqualDiff(expected, out)
        self.assertEqualDiff(err, "")



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.tests import TestCaseWithTransport


class TestCatRevision(TestCaseWithTransport):

    def test_cat_unicode_revision(self):
        tree = self.make_branch_and_tree('.')
        tree.commit('This revision', rev_id=b'abcd')
        output, errors = self.run_bzr(['cat-revision', 'abcd'])
        self.assertContainsRe(output, 'This revision')
        self.assertEqual('', errors)

    def test_cat_revision(self):
        """Test brz cat-revision.
        """
        wt = self.make_branch_and_tree('.')
        r = wt.branch.repository

        wt.commit('Commit one', rev_id=b'a@r-0-1')
        wt.commit('Commit two', rev_id=b'a@r-0-2')
        wt.commit('Commit three', rev_id=b'a@r-0-3')

        with r.lock_read():
            revs = {}
            for i in (1, 2, 3):
                revid = b"a@r-0-%d" % i
                stream = r.revisions.get_record_stream([(revid,)], 'unordered',
                                                       False)
                revs[i] = next(stream).get_bytes_as('fulltext')

        for i in [1, 2, 3]:
            self.assertEqual(revs[i],
                             self.run_bzr('cat-revision -r revid:a@r-0-%d' % i)[0].encode('utf-8'))
            self.assertEqual(revs[i],
                             self.run_bzr('cat-revision a@r-0-%d' % i)[0].encode('utf-8'))
            self.assertEqual(revs[i],
                             self.run_bzr('cat-revision -r %d' % i)[0].encode('utf-8'))

    def test_cat_no_such_revid(self):
        tree = self.make_branch_and_tree('.')
        err = self.run_bzr('cat-revision abcd', retcode=3)[1]
        self.assertContainsRe(
            err, 'The repository .* contains no revision abcd.')

    def test_cat_revision_directory(self):
        """Test --directory option"""
        tree = self.make_branch_and_tree('a')
        tree.commit('This revision', rev_id=b'abcd')
        output, errors = self.run_bzr(['cat-revision', '-d', 'a', 'abcd'])
        self.assertContainsRe(output, 'This revision')
        self.assertEqual('', errors)

    def test_cat_tree_less_branch(self):
        tree = self.make_branch_and_tree('.')
        tree.commit('This revision', rev_id=b'abcd')
        tree.controldir.destroy_workingtree()
        output, errors = self.run_bzr(['cat-revision', '-d', 'a', 'abcd'])
        self.assertContainsRe(output, 'This revision')
        self.assertEqual('', errors)



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for our estimation of compressed content."""

import hashlib
import zlib

from .. import estimate_compressed_size, tests


class TestZLibEstimator(tests.TestCase):

    def get_slightly_random_content(self, length, seed=b''):
        """We generate some hex-data that can be seeded.

        The output should be deterministic, but the data stream is effectively
        random.
        """
        h = hashlib.md5(seed)
        hex_content = []
        count = 0
        while count < length:
            b = h.hexdigest().encode('ascii')
            hex_content.append(b)
            h.update(b)
            count += len(b)
        return b''.join(hex_content)[:length]

    def test_adding_content(self):
        ze = estimate_compressed_size.ZLibEstimator(32000)
        raw_data = self.get_slightly_random_content(60000)
        block_size = 1000
        for start in range(0, len(raw_data), block_size):
            ze.add_content(raw_data[start:start + block_size])
            if ze.full():
                break
        # Practise showed that 'start' was 56000. However, zlib is a bit
        # platform dependent, so give it +/- 5%.
        self.assertTrue(54000 <= start <= 58000,
                        "Unexpected amount of raw data added: %d bytes" % (start,))
        # The real compression should be 'close' to 32000, real measurement was
        # 32401
        raw_comp = zlib.compress(raw_data[:start])
        self.assertTrue(31000 < len(raw_comp) < 33000,
                        "Unexpected compressed size: %d bytes" % (len(raw_comp),))

    def test_adding_more_content(self):
        ze = estimate_compressed_size.ZLibEstimator(64000)
        raw_data = self.get_slightly_random_content(150000)
        block_size = 1000
        for start in range(0, len(raw_data), block_size):
            ze.add_content(raw_data[start:start + block_size])
            if ze.full():
                break
        # Practise showed that 'start' was 112000
        self.assertTrue(110000 <= start <= 114000,
                        "Unexpected amount of raw data added: %d bytes" % (start,))
        # The real compression should be 'close' to 32000, real measurement was
        # 32401
        raw_comp = zlib.compress(raw_data[:start])
        self.assertTrue(63000 < len(raw_comp) < 65000,
                        "Unexpected compressed size: %d bytes" % (len(raw_comp),))



# Copyright (C) 2006, 2007, 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy/generate_ids.py"""

from ... import tests
from .. import generate_ids


class TestFileIds(tests.TestCase):
    """Test functions which generate file ids"""

    def assertGenFileId(self, regex, filename):
        """gen_file_id should create a file id matching the regex.

        The file id should be ascii, and should be an 8-bit string
        """
        file_id = generate_ids.gen_file_id(filename)
        self.assertContainsRe(file_id, b'^' + regex + b'$')
        # It should be a utf8 file_id, not a unicode one
        self.assertIsInstance(file_id, bytes)
        # gen_file_id should always return ascii file ids.
        file_id.decode('ascii')

    def test_gen_file_id(self):
        gen_file_id = generate_ids.gen_file_id

        # We try to use the filename if possible
        self.assertStartsWith(gen_file_id('bar'), b'bar-')

        # but we squash capitalization, and remove non word characters
        self.assertStartsWith(gen_file_id('Mwoo oof\t m'), b'mwoooofm-')

        # We also remove leading '.' characters to prevent hidden file-ids
        self.assertStartsWith(gen_file_id('..gam.py'), b'gam.py-')
        self.assertStartsWith(gen_file_id('..Mwoo oof\t m'), b'mwoooofm-')

        # we remove unicode characters, and still don't end up with a
        # hidden file id
        self.assertStartsWith(gen_file_id('\xe5\xb5.txt'), b'txt-')

        # Our current method of generating unique ids adds 33 characters
        # plus an serial number (log10(N) characters)
        # to the end of the filename. We now restrict the filename portion to
        # be <= 20 characters, so the maximum length should now be approx < 60

        # Test both case squashing and length restriction
        fid = gen_file_id('A' * 50 + '.txt')
        self.assertStartsWith(fid, b'a' * 20 + b'-')
        self.assertTrue(len(fid) < 60)

        # restricting length happens after the other actions, so
        # we preserve as much as possible
        fid = gen_file_id('\xe5\xb5..aBcd\tefGhijKLMnop\tqrstuvwxyz')
        self.assertStartsWith(fid, b'abcdefghijklmnopqrst-')
        self.assertTrue(len(fid) < 60)

    def test_file_ids_are_ascii(self):
        tail = br'-\d{14}-[a-z0-9]{16}-\d+'
        self.assertGenFileId(b'foo' + tail, 'foo')
        self.assertGenFileId(b'foo' + tail, 'foo')
        self.assertGenFileId(b'bar' + tail, 'bar')
        self.assertGenFileId(b'br' + tail, 'b\xe5r')

    def test__next_id_suffix_sets_suffix(self):
        generate_ids._gen_file_id_suffix = None
        generate_ids._next_id_suffix()
        self.assertNotEqual(None, generate_ids._gen_file_id_suffix)

    def test__next_id_suffix_increments(self):
        generate_ids._gen_file_id_suffix = b"foo-"
        generate_ids._gen_file_id_serial = 1
        try:
            self.assertEqual(b"foo-2", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-3", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-4", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-5", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-6", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-7", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-8", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-9", generate_ids._next_id_suffix())
            self.assertEqual(b"foo-10", generate_ids._next_id_suffix())
        finally:
            # Reset so that all future ids generated in the test suite
            # don't end in 'foo-XXX'
            generate_ids._gen_file_id_suffix = None
            generate_ids._gen_file_id_serial = 0

    def test_gen_root_id(self):
        # Mostly just make sure gen_root_id() exists
        root_id = generate_ids.gen_root_id()
        self.assertStartsWith(root_id, b'tree_root-')


class TestGenRevisionId(tests.TestCase):
    """Test generating revision ids"""

    def assertGenRevisionId(self, regex, username, timestamp=None):
        """gen_revision_id should create a revision id matching the regex"""
        revision_id = generate_ids.gen_revision_id(username, timestamp)
        self.assertContainsRe(revision_id, b'^' + regex + b'$')
        # It should be a utf8 revision_id, not a unicode one
        self.assertIsInstance(revision_id, bytes)
        # gen_revision_id should always return ascii revision ids.
        revision_id.decode('ascii')

    def test_timestamp(self):
        """passing a timestamp should cause it to be used"""
        self.assertGenRevisionId(
            br'user@host-\d{14}-[a-z0-9]{16}', 'user@host')
        self.assertGenRevisionId(b'user@host-20061102205056-[a-z0-9]{16}',
                                 'user@host', 1162500656.688)
        self.assertGenRevisionId(br'user@host-20061102205024-[a-z0-9]{16}',
                                 'user@host', 1162500624.000)

    def test_gen_revision_id_email(self):
        """gen_revision_id uses email address if present"""
        regex = br'user\+joe_bar@foo-bar\.com-\d{14}-[a-z0-9]{16}'
        self.assertGenRevisionId(regex, 'user+joe_bar@foo-bar.com')
        self.assertGenRevisionId(regex, '<user+joe_bar@foo-bar.com>')
        self.assertGenRevisionId(regex, 'Joe Bar <user+joe_bar@foo-bar.com>')
        self.assertGenRevisionId(regex, 'Joe Bar <user+Joe_Bar@Foo-Bar.com>')
        self.assertGenRevisionId(
            regex, 'Joe B\xe5r <user+Joe_Bar@Foo-Bar.com>')

    def test_gen_revision_id_user(self):
        """If there is no email, fall back to the whole username"""
        tail = br'-\d{14}-[a-z0-9]{16}'
        self.assertGenRevisionId(b'joe_bar' + tail, 'Joe Bar')
        self.assertGenRevisionId(b'joebar' + tail, 'joebar')
        self.assertGenRevisionId(b'joe_br' + tail, 'Joe B\xe5r')
        self.assertGenRevisionId(br'joe_br_user\+joe_bar_foo-bar.com' + tail,
                                 'Joe B\xe5r <user+Joe_Bar_Foo-Bar.com>')

    def test_revision_ids_are_ascii(self):
        """gen_revision_id should always return an ascii revision id."""
        tail = br'-\d{14}-[a-z0-9]{16}'
        self.assertGenRevisionId(b'joe_bar' + tail, 'Joe Bar')
        self.assertGenRevisionId(b'joe_bar' + tail, 'Joe Bar')
        self.assertGenRevisionId(b'joe@foo' + tail, 'Joe Bar <joe@foo>')
        # We cheat a little with this one, because email-addresses shouldn't
        # contain non-ascii characters, but generate_ids should strip them
        # anyway.
        self.assertGenRevisionId(b'joe@f' + tail, 'Joe Bar <joe@f\xb6>')
